/** \file main.c
 * \brief Sample STM32 project
 * \details This file holds a very basic code for STM32F103RB. This code
 * enables all GPIO ports, configures Flash wait-states and enables the PLL
 * to achieve the highest allowed frequency for STM32F103RB (72MHz). Main
 * code block just blinks the LED. The LED port and pin are defined in
 * config.h file. Target core frequency and quartz crystal resonator
 * frequency are defined there as well.
 *
 * \author Freddie Chopin, http://www.freddiechopin.info/
 * \date 2012-01-07
 */

/******************************************************************************
* project: WAV_SB16
* chip: STM32F103RB
* compiler: arm-none-eabi-gcc (Sourcery CodeBench Lite 2011.09-69) 4.6.1
*
* prefix: (none)
*
* available global functions:
* 	int main(void)
*
* available local functions:
* 	static void flash_latency(uint32_t frequency)
* 	static uint32_t pll_start(uint32_t crystal, uint32_t frequency)
* 	static void system_init(void)
*
* available interrupt handlers:
******************************************************************************/

/*
+=============================================================================+
| includes
+=============================================================================+
*/

#include <stdint.h>
#include <string.h>

#include "inc/stm32f10x.h"

#include "config.h"

#include "hdr/hdr_rcc.h"
#include "hdr/hdr_gpio.h"

#include "gpio.h"

#include "isa.h"
#include "sb16.h"

#include "fatfs/ff.h"
#include "fatfs/diskio.h"

#define FCC(c1,c2,c3,c4)	(((DWORD)c4<<24)+((DWORD)c3<<16)+((WORD)c2<<8)+(BYTE)c1)
#define BUFFER				1024

volatile uint8_t	page1[BUFFER],
					page2[BUFFER],
					buf = 0,
					buf_c = 0,
					bits = 0,
					channels = 0;
volatile uint16_t	n = 0;
volatile uint32_t	freq = 0;

FATFS		fatfs;
FRESULT		res;
FIL			file;
FILINFO		finfo;
DIR			dir;
static UINT	bytes;

/*
+=============================================================================+
| module variables
+=============================================================================+
*/

/*
+=============================================================================+
| local functions' declarations
+=============================================================================+
*/

static void flash_latency(uint32_t frequency);
static uint32_t pll_start(uint32_t crystal, uint32_t frequency);
static void system_init(void);

FRESULT Play(const char *, const char *);
DWORD CheckWAV(void);
void ReadData(void);

/*
+=============================================================================+
| global functions
+=============================================================================+
*/

/*------------------------------------------------------------------------*//**
* \brief main code block.
*//*-------------------------------------------------------------------------*/

int main(void)
{
	char *dir_;

	system_init();										//enable GPIO and AFIO
	pll_start(CRYSTAL, FREQUENCY);						//configure PLL
	SysTick_Config(FREQUENCY/800);						//configure systick for RTC function

	ISA_Init();
	SB_Init();

	for(;;)
	{
		if(f_mount(0, &fatfs) == FR_OK)					/* Initialize FS */
		{
			res = f_opendir(&dir, dir_ = "wav");		/* Open sound file directory */
			if(res == FR_NO_PATH)
			{
				res = f_opendir(&dir, dir_ = "");		/* Open root directory if no sound dir */
			}
			while(res == FR_OK)
			{
				res = f_readdir(&dir, 0);				/* Rewind dir */
				while(res == FR_OK)						/* Play all wav files in the dir */
				{
					res = f_readdir(&dir, &finfo);		/* Get a dir entry */
					if(res || !finfo.fname[0]) break;	/* Break on error or end of dir */
					if(!(finfo.fattrib & (AM_DIR|AM_HID)) && strstr(finfo.fname, ".WAV"))
						res = Play(dir_, finfo.fname);	/* Play file */
				}
			}
		}
	}
}

/*
+=============================================================================+
| local functions
+=============================================================================+
*/

/*------------------------------------------------------------------------*//**
* \brief Configures Flash latency
* \details Configures Flash latency (wait-states) which allows the chip to run
* at higher speeds
*
* \param [in] frequency defines the target frequency of the core
*//*-------------------------------------------------------------------------*/

static void flash_latency(uint32_t frequency)
{
	uint32_t wait_states;

	if (frequency < 24000000ul)				// 0 wait states for core speed below 24MHz
		wait_states = 0;
	else if (frequency < 48000000ul)		// 1 wait state for core speed between 24MHz and 48MHz
		wait_states = 1;
	else									// 2 wait states for core speed over 48MHz
		wait_states = 2;

	FLASH->ACR |= wait_states;				// set the latency
	return;
}

/*------------------------------------------------------------------------*//**
* \brief Starts the PLL
* \details Configure and enable PLL to achieve some frequency with some crystal.
* Before the speed change Flash latency is configured via flash_latency(). PLL
* parameter mul is based on both function parameters. The PLL is set up,
* started and connected. APB1 clock ratio is set to 1:2 (max freq = 36MHz)
*
* \param [in] crystal is the frequency of the crystal resonator connected to the
* STM32F103RB
* \param [in] frequency is the desired target frequency after enabling the PLL
*
* \return real frequency that was set
*//*-------------------------------------------------------------------------*/

static uint32_t pll_start(uint32_t crystal, uint32_t frequency)
{
	uint32_t mul;

	RCC_CR_HSEON_bb = 1;					// enable HSE clock
	flash_latency(frequency);				// configure Flash latency for desired frequency

	mul = frequency / crystal;				// PLL multiplier calculation

	if (mul > 16)							// max PLL multiplier is 16
		mul = 16;

	frequency = crystal * mul;

	RCC->CFGR |= ((mul - 2) << RCC_CFGR_PLLMUL_bit) | RCC_CFGR_PLLSRC | RCC_CFGR_PPRE1_DIV2;	// configuration of PLL: HSE x (mul), APB1 clk = /2

	while (!RCC_CR_HSERDY_bb);				// wait for stable clock

	RCC_CR_PLLON_bb = 1;					// enable PLL
	while (!RCC_CR_PLLRDY_bb);				// wait for PLL lock

	RCC->CFGR |= RCC_CFGR_SW_PLL;			// change SYSCLK to PLL
	while (((RCC->CFGR) & RCC_CFGR_SWS) != RCC_CFGR_SWS_PLL);	// wait for switch

	return frequency;
}

/*------------------------------------------------------------------------*//**
* \brief Initializes system
* \details Enables all GPIO ports
*//*-------------------------------------------------------------------------*/

static void system_init(void)
{
	gpio_init();
	return;
}

//=============================================================================

FRESULT Play(
	const char *dirp,								/* Directory */
	const char *fnp									/* File */
)
{
	DWORD check;
	strcpy((char*)page1, dirp);						/* sprintf(Buff, "%s/%s", dir, fn); */
	strcat((char*)page1, "/");
	strcat((char*)page1, fnp);
	res = f_open(&file, (char*)page1, FA_READ);
	n = buf = buf_c = 0;
	check = CheckWAV();								//check .wav file for information of sampling, bit depth, etc.
	if(check == 4 || check == 0)
		return 0;									//can't play this file
	ReadData();										//read first data into buffer
	SB_DSP_Reset();
	SB_DSP_SetTimeConst(freq);
	SB_DSP_DMA_Config(bits, channels, BUFFER);
	while(f_eof(&file) == FALSE && bytes == BUFFER)
	{
		if(buf_c)
		{
			ReadData();								//read data to the buffer
			buf_c = 0;
		}
	}
	while(n < bytes);
	SB_DSP_DMA_Close(bits);
	res = f_close(&file);							//close opened file
	return 0;
}

DWORD CheckWAV(void)
{
	BYTE al = 0;
	DWORD size = 0;

	if(f_read(&file, &page1, 12, &bytes)) return 0;
	if(bytes != 12 || LD_DWORD(page1+8) != FCC('W','A','V','E')) return 0;

	while(1)
	{
		f_read(&file, &page1, 8, &bytes);						/* Get Chunk ID and size */
		if (bytes != 8) return 0;
		size = LD_DWORD(&page1[4]);								/* Chunk size */

		switch(LD_DWORD(&page1[0]))
		{	/* FCC */
		case FCC('f','m','t',' '):								/* 'fmt ' chunk */
			if(size & 1) size++;								/* Align chunk size */
			if(size > 100 || size < 16) return 0;				/* Check chunk size */
			f_read(&file, &page1, size, &bytes);				/* Get content */
			if(bytes != size) return 0;
			if(page1[0] != 1) return 0;							/* Check coding type (LPCM) */
			if(page1[2] != 1 && page1[2] != 2) return 0;		/* Check channels (1/2) */
			channels = al = page1[2];							/* Save channel flag */
			if(page1[14] != 8 && page1[14] != 16) return 0;		/* Check resolution (8/16 bit) */
			bits = page1[14];									/* Save resolution flag */
			if(page1[14] & 16) al <<= 1;
			freq = LD_DWORD(&page1[4]);							/* Sampling freq */
			if(freq < 5000 || freq > 44100) return 4;			/* Check sampling freqency (5k-44.1k) */
			break;

		case FCC('d','a','t','a'):								/* 'data' chunk */
			if(!al) return 0;									/* Check if format valid */
			if(size < 1024 || (size & (al - 1))) return 0;		/* Check size */
			if(f_tell(&file) & (al - 1)) return 0;				/* Check offset */
			return size;										/* Start to play */

		case FCC('D','I','S','P'):								/* 'DISP' chunk */
		case FCC('L','I','S','T'):								/* 'LIST' chunk */
		case FCC('f','a','c','t'):								/* 'fact' chunk */
			if(size & 1) size++;								/* Align chunk size */
			f_lseek(&file, f_tell(&file) + size);				/* Skip this chunk */
			break;

		default:												/* Unknown chunk */
			return 0;
			break;
		}
	}
	return 0;
}

void ReadData(void)
{
	if(buf == 0)
	{
		res = f_read(&file,&page1,BUFFER,&bytes);			//read data to first buffer
	}else{
		res = f_read(&file,&page2,BUFFER,&bytes);			//read data to second buffer
	}
	return;
}

/*
+=============================================================================+
| ISRs
+=============================================================================+
*/

void SysTick_Handler(void)
{
	disk_timerproc();
	return;
}

void EXTI9_5_IRQHandler(void)
{
	CLR_ALE;
	SET_IOW;
	SET_IOR;
	if(bits == 16)
		ISA_Read_IO(BASE_ADDR|0xF);
	else if(bits == 8)
		ISA_Read_IO(BASE_ADDR|0xE);
	n = 0;
	buf_c = 1;
	buf = ~buf;
	EXTI->PR |= (1<<IRQ5_PIN);
	return;
}

void EXTI1_IRQHandler(void)
{
	CLR_ALE;
	SET_IOW;
	SET_IOR;
	CLR_DACK0;
	if(buf != 0)
	{
		ISA_DMA_Write_IO(BASE_ADDR|0xC,page1[n]);
	}else{
		ISA_DMA_Write_IO(BASE_ADDR|0xC,page2[n]);
	}
	n += 1;
	SET_DACK0;
	EXTI->PR |= (1<<DRQ0_PIN);
	return;
}

void EXTI2_IRQHandler(void)
{
	CLR_ALE;
	SET_IOW;
	SET_IOR;
	CLR_DACK5;
	if(buf != 0)
	{
		ISA_DMA_Write_IO(BASE_ADDR|0xC,(page1[n+1]<<8)|page1[n]);
	}else{
		ISA_DMA_Write_IO(BASE_ADDR|0xC,(page2[n+1]<<8)|page2[n]);
	}
	n += 2;
	SET_DACK5;
	EXTI->PR |= (1<<DRQ5_PIN);
	return;
}

/******************************************************************************
* END OF FILE
******************************************************************************/
